@@ -6,6 +6,7 @@ import logging
 import re
 import weakref
 from lxml import etree
+from xml.etree import ElementTree as XmlET
 
 HERE = os.path.abspath(os.path.dirname(__file__))
 XSD_PATH = os.path.join(HERE, 'resources', 'SCL_Schema', 'SCL.xsd')
@@ -22,6 +23,8 @@ REF_SCL_NODES = r'(?:\{.+\})?(?:Header|Substation|Private|Communication)'
 SEP1 = '$'          # Standard MMS separator.
 SEP2 = '/'          # MMS Separator for system testing
 
+FORCE_ITER_MODE = False
+
 NODES_ATTRS = {
     'IED': [
         'Server',
@@ -103,15 +106,6 @@ NODES_ATTRS = {
         'fc',
         'ix'
     ],
-    'SubNetWork' : [
-        'type',
-        'name'
-    ],
-    'ConnectedAP' : [
-        'iedName',
-        'apName',
-        'redProt'
-    ],
 }
 
 
@@ -162,16 +156,16 @@ def _get_node_name(node: etree.Element):
             inst = node.get('inst')
             name = lnClass + str(inst)
         elif tag == 'LDevice':
-            name = node.get('ldName')
+            name = node.get('inst')
         elif tag == 'Private':
             name = node.get('type')
         elif tag == 'ConnectedAP':
-            name = node.get('iedName')        
+            name = node.get('iedName')
         else:
             name = tag
 
     if name is not None:
-        name = re.sub('[-:]', '_', name)  # Pour les RTE-FIP, RTE-BAP, rte:BAP et rte:FIP
+        name = re.sub('[-:]', '_', name)  # in to handle: RTE-FIP, RTE-BAP, rte:BAP and rte:FIP
 
     return name
 
@@ -209,6 +203,27 @@ class DataTypeTemplates:
         item_xpath = 'child::*[@id="{}"]'.format(id)
         return self._datatypes_root.xpath(item_xpath, namespaces=NS)[0]
 
+    def get_Data_Type_Definitions(self) -> dict:
+
+        """
+            Create a table of all Data Types definitions (LNodeType, DOType, DAType ad EnumType)
+
+            Returns
+
+            -------
+
+            `dict`
+
+                A dictionnary of the DataTypes definitions grouped by tag
+        """
+        tags = {'LNodeType':[], 'DOType':[], 'DAType':[], 'EnumType':[]}
+
+        for tag_key in tags.keys():
+            item_xpath = 'child::iec61850:{}'.format(tag_key)
+            tags[tag_key] = self._datatypes_root.xpath(item_xpath, namespaces=NS)
+
+        return tags
+
 
 # TODO :
 # - Manage BAP/FIP
@@ -218,6 +233,10 @@ class SCDNode:
     """
         Basic class to compute SCD nodes
     """
+    @property
+    def children(self):
+        return self.get_children()
+
     def __init__(self, datatypes: DataTypeTemplates, node_elem: etree.Element = None, fullattrs: bool = False, **kwargs: dict):
         """
             Constructor
@@ -314,10 +333,8 @@ class SCDNode:
             new_node = LN(self._datatypes, elem, self._fullattrs, **attributes)
         elif elem.tag.split('}')[-1] == 'LN0':
             new_node = LN0(self._datatypes, elem, self._fullattrs, **attributes)
-        elif elem.tag.split('}')[-1] == 'SubNetwork':
-            new_node = SubNetwork(self._datatypes, elem, self._fullattrs, **attributes)
-        elif elem.tag.split('}')[-1] == 'ConnectedAP':
-            new_node = ConnectedAP(self._datatypes, elem, self._fullattrs, **attributes)
+        elif elem.tag.split('}')[-1] == 'LDevice':
+            new_node = LD(self._datatypes, elem, self._fullattrs, **attributes)
         else:
             new_node = SCDNode(self._datatypes, elem, self._fullattrs, **attributes)
 
@@ -332,7 +349,7 @@ class SCDNode:
 
         return new_node
 
-    def get_DA_leaf_nodes(self) -> {}:
+    def get_DA_leaf_nodes(self) -> dict:
         """
             Recursively retrieve the leaf DA nodes of the current node.
 
@@ -350,7 +367,7 @@ class SCDNode:
         self._collect_DA_leaf_nodes(self, leaves, mms)
         return leaves
 
-    def get_children(self, tag: str = None) -> []:
+    def get_children(self, tag: str = None) -> list:
         """
             Retrieve the children nodes of the current node.
             (not recursive)
@@ -448,7 +465,7 @@ class SCDNode:
 
         return len(node.get_children()) == 0 and isinstance(node, DA) and hasattr(node, 'parent')
 
-    def _collect_DA_leaf_nodes(self, node, leaves: {}, mms: bool = False) -> {}:
+    def _collect_DA_leaf_nodes(self, node, leaves: dict, mms: bool = False) -> dict:
         """
             /!\\ PRIVATE : do not use /!\\
 
@@ -698,9 +715,7 @@ class LN(SCDNode):
         self._all_attributes.extend(NODES_ATTRS['LN'])
 
         if node_elem is not None:
-            lnClass = node_elem.get('lnClass')
-            inst = node_elem.get('inst')
-            kwargs['name'] = lnClass + str(inst)
+            kwargs['name'] = _get_node_name(node_elem)
         else:
             lnClass = kwargs.get('lnClass')
             inst = kwargs.get('inst')
@@ -772,11 +787,10 @@ class LD(SCDNode):
         self._all_attributes.extend(NODES_ATTRS['LD'])
 
         if node_elem is not None:
-            ldName = node_elem.get('ldName')
-            kwargs['name'] = ldName
+            kwargs['name'] = _get_node_name(node_elem)
         else:
-            ldName = kwargs.get('ldName')
-            kwargs.update({'name': ldName})
+            inst = kwargs.get('inst')
+            kwargs.update({'name': inst})
 
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
@@ -811,68 +825,6 @@ class IED(SCDNode):
         super().__init__(datatypes, node_elem, fullattrs, **kwargs)
 
 
-class SubNetwork(SCDNode):
-    """
-        Class to manage an SubNetwork
-    """
-
-    def __init__(self, datatypes: DataTypeTemplates, node_elem: etree.Element = None, fullattrs=False, **kwargs: dict):
-        """
-            Constructor
-
-            Parameters
-            ----------
-            `datatypes`
-                Instance of the DataTypeTemplates object from the SCD/SCL file.
-
-            `node_elem` (optional)
-                etree.Element element from the SCD/SCL file to build the node object.
-
-            `fullattrs` (optional)
-                If True, all the possible attributes for the SCD objects will be created
-                even if they are not described in the SCD/SCL file.
-
-            `kwargs` (optional)
-                Dictionary of the node attributes.
-
-            /!\\ At least one of node_elem or kwargs must be provided /!\\
-        """
-        self._all_attributes = []
-        self._all_attributes.extend(NODES_ATTRS['SubNetWork'])
-        super().__init__(datatypes, node_elem, fullattrs, **kwargs)
-
-
-class ConnectedAP(SCDNode):
-    """
-        Class to manage an IED
-    """
-
-    def __init__(self, datatypes: DataTypeTemplates, node_elem: etree.Element = None, fullattrs=False, **kwargs: dict):
-        """
-            Constructor
-
-            Parameters
-            ----------
-            `datatypes`
-                Instance of the DataTypeTemplates object from the SCD/SCL file.
-
-            `node_elem` (optional)
-                etree.Element element from the SCD/SCL file to build the node object.
-
-            `fullattrs` (optional)
-                If True, all the possible attributes for the SCD objects will be created
-                even if they are not described in the SCD/SCL file.
-
-            `kwargs` (optional)
-                Dictionary of the node attributes.
-
-            /!\\ At least one of node_elem or kwargs must be provided /!\\
-        """
-        self._all_attributes = []
-        self._all_attributes.extend(NODES_ATTRS['ConnectedAP'])
-        super().__init__(datatypes, node_elem, fullattrs, **kwargs)
-
-
 class SCD_handler():
     """
         Class to handle a SCD/SCL file
@@ -890,6 +842,7 @@ class SCD_handler():
                 If True, all the possible attributes for the SCD objects will be created
                 even if they are not described in the SCD/SCL file.
         """
+        self._scl_root = None
         self._scd_path = scd_path
         schema_doc = etree.parse(XSD_PATH)
         self._schema = etree.XMLSchema(schema_doc)
@@ -902,13 +855,60 @@ class SCD_handler():
             raise AttributeError(err)
 
         self.datatypes = DataTypeTemplates(self._scd_path)
-
-        scl_children = self._iter_get_SCL_elems()
+        self.Substation = []
+        scl_children = self._get_SCL_elems()
         for elem in scl_children:
-            elem_name = _get_node_name(elem)
-            setattr(self, elem_name, SCDNode(self.datatypes, elem, self._fullattrs))
+            if elem.tag.split('}')[-1] == 'Substation':
+                self.Substation.append(SCDNode(self.datatypes, elem, self._fullattrs))
+            else:
+                elem_name = _get_node_name(elem)
+                setattr(self, elem_name, SCDNode(self.datatypes, elem, self._fullattrs))
 
-    def get_all_IEDs(self) -> [IED]:
+    def extract_sub_SCD(self, ied_name_list: list) -> str:
+        """
+            Extract a smaller SCD containing only the IED from the IED_list
+
+            Parameters
+            ----------
+            `ied_name_list`
+                List of the ied names to keep
+
+            Returns
+            -------
+            `str`
+                The full path of the new reduced scd file
+        """
+        newroot = None
+        path, ext = os.path.splitext(self._scd_path)
+        dest_path = '{}-reduced{}'.format(path, ext)
+
+        ctx = etree.iterparse(self._scd_path, events=("start",), tag='{}SCL'.format(SCL_NAMESPACE))
+        for _, root in ctx:
+            newroot = etree.Element(root.tag, nsmap=root.nsmap, attrib=root.attrib)
+            break
+
+        scl_children = self._get_SCL_elems()
+        for child in scl_children:
+            newroot.append(child)
+
+        #Clean not needed ConnectedAP
+        connected_aps = newroot.xpath('child::iec61850:Communication/*/iec61850:ConnectedAP', namespaces=NS)
+        connected_aps = [itm for itm in connected_aps if itm.get('iedName') not in ied_name_list]
+        for c_ap in connected_aps:
+            c_ap.getparent().remove(c_ap)
+
+        ieds = self._get_IED_elems_by_names(ied_name_list)
+        for ied in ieds:
+            newroot.append(ied)
+
+        newroot.append(self.datatypes._datatypes_root)
+
+        et = etree.ElementTree(newroot)
+        et.write(dest_path, encoding="utf-8", xml_declaration=True)
+        
+        return dest_path
+
+    def get_all_IEDs(self) -> list:
         """
             Load all the IEDs from the SCD/SCL file
 
@@ -917,9 +917,14 @@ class SCD_handler():
             `[IED]`
                 An array of the loaded IED objects
         """
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        ieds = []
+        if self._scl_root is not None:
+            ieds = self._scl_root.xpath('child::iec61850:IED', namespaces=NS)
+        else:
+            ieds = self._iter_get_all_IEDs()
+
         tIED = []
-        for _, ied in context:
+        for ied in ieds:
             tIED.append(IED(self.datatypes, ied, self._fullattrs))
 
         return tIED
@@ -938,15 +943,11 @@ class SCD_handler():
             `IED`
                 The loaded IED object
         """
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
-        for _, ied in context:
-            item_name = ied.get('name')
-            if item_name == ied_name:
-                return IED(self.datatypes, ied, self._fullattrs)
-            else:
-                ied.clear()
+        ied_elems = self._get_IED_elems_by_names([ied_name])
+        if len(ied_elems) > 0:
+            return IED(self.datatypes, ied_elems[0], self._fullattrs)
 
-    def get_IED_names_list(self) -> [str]:
+    def get_IED_names_list(self) -> list:
         """
             Load an IED from the SCD/SCL file by name
 
@@ -956,14 +957,16 @@ class SCD_handler():
                 The found IED names
         """
         result = []
-        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
-        for _, ied in context:
-            result.append(ied.get('name'))
-            ied.clear()
+        if self._scl_root is not None:
+            ieds = self._scl_root.xpath('child::iec61850:IED', namespaces=NS)
+            for ied in ieds:
+                result.append(ied.get('name'))
+        else:
+            result = self._iter_get_IED_names_list()
 
         return result
-
-    def _check_scd_file(self) -> (bool, str):
+    
+    def _check_scd_file(self) -> tuple:
         """
             /!\\ PRIVATE : do not use /!\\
 
@@ -983,13 +986,39 @@ class SCD_handler():
         if file_size < MAX_VALIDATION_SIZE:
             tree = etree.parse(self._scd_path)
             is_valid = self._schema.validate(tree)
-            tree = None
+            if not FORCE_ITER_MODE:
+                self._scl_root = tree.getroot()
             return (is_valid, self._schema.error_log.last_error)
         else:
             LOGGER.warn('XSD validation skipped due to file size over {} Mo' % MAX_VALIDATION_SIZE)
             return True
 
-    def _iter_get_SCL_elems(self) -> [etree.Element]:
+    def _get_IED_elems_by_names(self, ied_names_list: list) -> list:
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Retrieve the IED elements by name
+
+            Parameters
+            ----------
+            `ied_names_list`
+                List of ied names to retrieve
+
+            Returns
+            -------
+            `[etree.Element]`
+                Array of found etree.Element elements
+        """
+        result = []
+        if self._scl_root is not None:
+            result = self._scl_root.xpath('child::iec61850:IED', namespaces=NS)
+            result = [ied for ied in result if ied.get('name') in ied_names_list]
+        else:
+            result = self._iter_get_IED_elems_by_names(ied_names_list)
+
+        return result
+
+    def _get_SCL_elems(self) -> list:
         """
             /!\\ PRIVATE : do not use /!\\
 
@@ -1001,28 +1030,47 @@ class SCD_handler():
             `[etree.Element]`
                 Array of found etree.Element elements
         """
-        tags = [
-            '{}Header'.format(SCL_NAMESPACE),
-            '{}Communication'.format(SCL_NAMESPACE),
-            '{}Substation'.format(SCL_NAMESPACE),
-            '{}Private'.format(SCL_NAMESPACE)
-        ]
-        context = etree.iterparse(self._scd_path, events=("end",), tag=tags)
-        result = []
-        for _, elem in context:
+        tags = ['Header', 'Communication', 'Substation', 'Private']
 
-            if elem.tag.split('}')[-1] != 'Private':
-                result.append(elem)
-            elif elem.tag.split('}')[-1] == 'Private' \
-                                            and (elem.xpath('following-sibling::iec61850:Header', namespaces=NS)
-                                            or elem.xpath('preceding-sibling::iec61850:Header', namespaces=NS)):
-                result.append(elem)
-            else:
-                elem.clear()
+        xpath_tags = ''
+        for tag in tags:
+            xpath_tags = '{} or self::iec61850:{}'.format(xpath_tags, tag)
+        
+        xpath_tags = xpath_tags[3:]
+
+        result = []
+        if self._scl_root is not None:
+            result = self._scl_root.xpath('child::*[{}]'.format(xpath_tags), namespaces=NS)
+        else:
+            result = self._iter_get_SCL_elems(tags)
 
         return result
 
-    def _iter_get_all_elem_by_tag(self, tag:str) -> [etree.Element]:
+    def _get_all_elem_by_tag(self, tag:str) -> list:
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Retrieve the SCL elements by tag
+
+            Parameters
+            ----------
+            `tag`
+                the elements tag to search
+
+            Returns
+            -------
+            `[etree.Element]`
+                Array of found etree.Element elements
+        """
+        elem_list = []
+        if self._scl_root is not None:
+            elem_list = self._scl_root.xpath('iec61850:{}'.format(tag), namespaces=NS)
+        else:
+            elem_list = self._iter_get_all_elem_by_tag(tag)
+
+        return elem_list
+
+    def _iter_get_all_elem_by_tag(self, tag:str) -> list:
         """
             /!\\ PRIVATE : do not use /!\\
 
@@ -1047,3 +1095,98 @@ class SCD_handler():
             _, elem = next(context)
 
         return elem_list
+
+    def _iter_get_SCL_elems(self, tags: list) -> list:
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Retrieve the SCL node direct children
+            Header, Communication, Substation, Private nodes
+
+            Parameters
+            ----------
+            `tags`
+                List of tags to consider
+
+            Returns
+            -------
+            `[etree.Element]`
+                Array of found etree.Element elements
+        """
+        for idx, tag in enumerate(tags):
+            tags[idx] = '{}{}'.format(SCL_NAMESPACE, tag)
+
+        context = etree.iterparse(self._scd_path, events=("end",), tag=tags)
+        result = []
+        for _, elem in context:
+
+            if elem.tag.split('}')[-1] != 'Private':
+                result.append(elem)
+            elif elem.tag.split('}')[-1] == 'Private' \
+                                            and (elem.xpath('following-sibling::iec61850:Header', namespaces=NS)
+                                            or elem.xpath('preceding-sibling::iec61850:Header', namespaces=NS)):
+                result.append(elem)
+            else:
+                elem.clear()
+
+        return result
+
+    def _iter_get_IED_elems_by_names(self, ied_names_list: list) -> list:
+        """
+            /!\\ PRIVATE : do not use /!\\
+
+            Retrieve the IED elements by name
+
+            Parameters
+            ----------
+            `ied_names_list`
+                List of ied names to retrieve
+
+            Returns
+            -------
+            `[etree.Element]`
+                Array of found etree.Element elements
+        """
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        result = []
+        for _, ied in context:
+            item_name = ied.get('name')
+            if item_name in ied_names_list:
+                result.append(ied)
+            else:
+                ied.clear()
+
+        return result
+
+    def _iter_get_IED_names_list(self) -> list:
+        """
+            Load an IED from the SCD/SCL file by name
+
+            Returns
+            -------
+            `[str]`
+                The found IED names
+        """
+        result = []
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        for _, ied in context:
+            result.append(ied.get('name'))
+            ied.clear()
+
+        return result
+
+    def _iter_get_all_IEDs(self) -> list:    
+        """
+            Load all the IEDs from the SCD/SCL file
+
+            Returns
+            -------
+            `[IED]`
+                An array of the loaded IED objects
+        """
+        context = etree.iterparse(self._scd_path, events=("end",), tag='{}IED'.format(SCL_NAMESPACE))
+        tIED = []
+        for _, ied in context:
+            tIED.append(ied)
+
+        return tIED
